;-*- Mode: Common-lisp; Package: langutils; Readtable: ytools; -*-
(in-package :langutils)
;;; $Id: namespace.lisp,v 1.24 2005/11/07 18:37:44 dvm Exp $

;;; Copyright (C) 1976-2003 
;;;     Drew McDermott and Yale University.  All rights reserved
;;; This software is released under the terms of the Modified BSD
;;; License.  See file COPYING for details.

(depends-on %module/ ytools)

(depends-on :at-run-time %langutils/ defectexp)

;;; A namespace is a set of associations between prefixed symbols and
;;; canonical symbols.  Sometimes the object associated is not a 
;;; Canonical-symbol, but a Lisp Symbol.

;;; Prefixed symbol readable and printed as
;;; @pref1:pref2:...:prefK:sym.  Treat pref1, pref2, ..., sym as strings.
;;; If you start in namespace N0, this means the canonical symbol 
;;; associated with sym in the space(s) found by the following rules:
;;; If K=0, include space N0.
;;; If N0 has parent N1 with prefix "", include the spaces obtained by
;;;   recursively applying the rules to sym and pref1, pref2,...,prefK
;;;   starting in N1.
;;; If K>1, and N0 has parent N1 with prefix pref1, include the spaces
;;;   obtained by recursively applying the rules to sym and pref2,...,prefK 
;;;   starting in N1.
;;; The rules can find multiple spaces, but they must all agree on the
;;; canonical symbol associated with the prefixed symbol.   
;;; The entry points to the abstract data type enforce this requirement.
;;; Note that there can be several parents with the same prefix, and
;;; several parents with different prefixes and the same namespace.
;;; But the prefix + the namespace can have at most one entry in the
;;; parents list.

;;; This mechanism is designed to supplant the Lisp package system, not
;;; exist on top of it.  When it's expedient to represent strings as symbols,
;;; they are always represented as keywords.

;;; Clarification: Not all symbols need to be canonicalized.  In Opt,
;;; the only symbols that are canonicalized are those with Domain scope.
;;; Those with global or local scope remain ordinary symbols (which are,
;;; of course, located in packages).

(eval-when (:compile-toplevel :load-toplevel :execute :slurp-toplevel)
   (export '(+namespace-mode+ +all-symbols+ +all-symbols-but+
	     namespace* place-Namespace Namespace-name Namespace-parents
	     new-Namespace-import Namespace-import-exporter
	     Namespace-exports place-Namespace-with-exports
	     Prefixed-symbol make-Prefixed-symbol is-Prefixed-symbol
	     Canonized-symbol
	     is-Canonized-symbol Canonized-symbol-sym Canonized-symbol-space
	     is-CSymbol
	     namespace-associate prefixed-symbol-canonicalize
	     namespace-has-import namespace-add-import
	     sym-ns-resolve exp-namespace-symbols-place
	     is-symbolish no-namespace-add namespace-show
	     local-name local-sym symbol-as-keyword namespace-change-exports)))

(debuggable 1)

(defvar +all-symbols+ ':all)
;;;;      (make-Printable (\\ (srm) (out (:to srm) (:a "+export_all+")))))
(defvar +all-symbols-but+ ':all-but)

;;; If this is false, the system behaves exactly like it did before Namespaces
;;; were introduced.  So it should be set here and not changed.  (Recompile
;;; to get old system.)
(defconstant +namespace-mode+ #.yt::lisplang-namespace-mode*)

;;; If +namespace-mode+ is true, then the following variable changes fundamentally
;;; the way relativized symbols are read and printed.--
(defvar namespace* false)
;;; -- Be very careful about its value at all times.  
;;; If it is false, then Prefixed-symbols are read as themselves, and
;;; Canonized-symbols are printed as themselves.
;;; If it is true, then things work more like the Lisp package system.
;;; A Prefixed-symbol P is read as the Canonized-symbol P denotes.
;;; A Canonized symbol C is printed as the Prefixed-symbol that would
;;; be read as C.
;;; It should definitely be false during 'opt-load' because parsers
;;; should control how names are allocated to different namespaces.

(def-class Namespace
   (:options :key)
   name
   (generation 0)
   (uri false) ;; some globally unique designator for this namespace.
   (table (make-hash-table :test #'eq) :type hash-table)
   ;; -- Associates symbols with objects.  Used only for symbols whose
   ;; canonical "home" is here.  In particular, the table is not used
   ;; as a cache for all namespace-lookup's; just those that end here.
   ;; If table is false, no symbols are cached here.
   (exports '())
   ;; -- List of symbols or +all-symbols+ or (+all-symbols-but+ -syms-)
   (parents '())
   ;; -- List of Namespace-imports (see below)
   (children '())
   ;; -- List of Namespaces that import from this one
   ;; 'parents' and 'children' must not be set directly;
   ;; instead use Namespace-add-import, defined below.
)

(def-meth
      print-object ((ns Namespace) srm)
	 (out (:to srm) "#<Namespace " (Namespace-name ns)
	      (:q ((Namespace-uri ns)
		   1 (Namespace-uri ns)))
	      ">"))

(defvar namespaces* (make-hash-table :test #'eq :size 100))

(defun place-Namespace-with-exports (name tab exports)
   (let ((ns (place-Namespace name tab true)))
      (!= (Namespace-exports ns) exports)
      ns))

(defun place-Namespace (name tab clear)   ;;;&optional (clear false clear-supplied))
   (let ((prev (href (or tab namespaces*) name false)))
      (cond (prev
	     (cond (clear
		    (!= (Namespace-generation prev) 0)
		    (!= (Namespace-exports prev) !())
		    (repeat :for ((par :in (Namespace-parents prev)))
		       (!= (Namespace-children (Namespace-import-exporter par))
			   (delete prev *-*)))
		    (!= (Namespace-parents prev) !())
		    (!= (Namespace-exports prev) !())
		    (clrhash (Namespace-table prev))))
	     prev)
	    (t
	     (let ((new-ns (make-Namespace :name name))
		   (name-uri (get name 'uri)))
		(cond (name-uri
		       (let ((net.uri-pkg (find-package :net.uri)))
			  (cond (net.uri-pkg
				 (!= (Namespace-uri new-ns) name)
				 (!= (Namespace-name new-ns) 
				     (funcall (intern :uri-fragment net.uri-pkg)
					      name)))
				(t
				 (signal-problem place-Namespace
				    "Attempt to use nonexistent URI facility for"
				    " namespace " name))))))
		(!= (href (or tab namespaces*) name)
		    new-ns)
		new-ns)))))

(defun find-namespace (name tab)
   (href (or tab namespaces*) name false))

(defmethod make-load-form ((ns Namespace) &optional env)
                          (declare (ignore env))
   (cond ((find-namespace (Namespace-name ns) namespaces*)
	  `(place-Namespace ',(Namespace-name ns) false false))
	 (t
	  (signal-problem make-load-form/Namespace
	     "No load form for non-global namespace: " ns))))   

;;; We say "namespace N is productive" to mean that, with respect to the
;;; "current" namespace, one can get to N via a series of productive
;;; Namespace-imports.

(def-class Namespace-import
   (:options (:medium :list))
   productive				; Boolean
   ;; -- if true, prefix must not be shared by any other import,
   ;; and a previously unseen symbol with this prefix will be created,
   ;; not generate an error message.
   exporter  ; a Namespace
   (prefix false :type (or string null))
   symbols)
;;; -- If 'symbols' is +all-symbols+, then take all of them from the corresponding
;;; parent.  If it's a list (+all-symbols-but+ -prefixed-syms-), 
;;; then take all but the syms.
;;; Otherwise, it's a list of Prefixed-symbols.
;;; The specified Prefixed-symbols are intersected with 
;;; those exported from 'exporter' to get the symbols that are actually
;;; inherited by the Namespace this import occurs as a parent of.

(defun new-Namespace-import (prod exporter prefix symbols)
   (make-Namespace-import
      prod exporter
      (cond ((equal prefix "") false)
	    (t prefix))
      symbols))

(defun Namespace-import-prefix-is-empty (imp)
   (let ((pref (Namespace-import-prefix imp)))
      (or (not pref) (string= pref ""))))

(defun namespace-import-prefix-cons (imp l)
   (let ((pref (Namespace-import-prefix imp)))
      (cond ((or (not pref)
		 (string= pref ""))
	     l)
	    (t (cons pref l)))))

(defvar sym-delims* '(#\Space #\Tab #\Newline #\( #\) #\" #\'
		      #\, #\;))

(def-class Prefixed-symbol
    prefix
    sym
   (:handler
      (print-object ((ps Prefixed-symbol) srm)
	 (let-fun ()
	    (out (:to srm)
	       (:e (:stream osrm)
		  (:o "@" (:e (repeat :for ((p :in (Prefixed-symbol-prefix ps)))
				 (substring-print p osrm) (:o ":")))
		       (:e (substring-print (symbol-name (Prefixed-symbol-sym ps))
					    osrm)))))
	  :where
	    (:def substring-print (ss osrm)
	        (repeat :for ((i = 0 :to (- (length ss) 1)))
		 :result (out (:to osrm) (:a (build-symbol (:package false) (:< ss))))
		 :until (member (elt ss i) sym-delims* :test #'char=)
		 :result (out (:to osrm) "\|" (:a ss) "\|")))))))

(cond ((-- +namespace-mode+)
       (set-macro-character
           '#\@
           (\\ (srm _)
              (let ((subs (read-syms-colon-sepped srm)))
                 (cond ((null subs)
                        (signal-problem \@-reader
                           "No substrings after @")))
                 (let ((prefixes (butlast subs))
                       (sym (bind ((*package* keyword-package*))
                               (read-from-string (lastelt subs)))))
                    (cond (namespace*
                           (or (namespace-lookup namespace* sym prefixes false)
                               (make-Prefixed-symbol prefixes sym)))
                          (t
                           (make-Prefixed-symbol prefixes sym)))))))))

;;; Read following characters as colon-separated symbol names,
;;; and return as list of strings.
;;; Escape characters keep whitespace from terminating, but otherwise
;;; just get included in the string unchanged.
;;; Escape characters make no sense before the last segment,
;;; but we overlook that here, so we don't have to think about whether
;;; the next segment is also the last.
(defun read-syms-colon-sepped (srm)
   (repeat :for ((found-end false)
		 :collectors segs)
      ;; Get next substring
    :collect
      (let ((cll (extract-to-colon)))
	 (coerce
            (<< nconc cll)
	    'string))
    :until found-end
    :result segs

    :where	 

      ;;; Returns a list of lists of chars.  Normally each list
      ;;; has exactly one char, but vertical bars can give rise
      ;;; to lists with other lengths.
      (:def extract-to-colon ()
	 (repeat :for ((ch = (in (:from srm) :peek)
			   :then :again)
		       (escaped-char = false :then :again)
		       :collector next-seg)
;;;;	    (out "(" ch ")")
	    (cond ((char= ch #\\)
		   (!= ch (in (:from srm) :char))
		   (!= escaped-char true)))
;;;;	    (out "_" (:a ch))
	    (!= found-end
		(or (eq ch eof*)
		    (and (not escaped-char)
			 (member ch sym-delims*
				 :test #'char=))))
	  :until (or found-end (char= ch #\:))
	  :result (progn (cond ((not found-end) (bite)))
			 (list next-seg))
	  :until (char= ch #\|)
	  :result (progn (bite)
			 (cons next-seg (extract-to-vert-bar)))
	  :collect (in (:from srm) :char)))

      (:def extract-to-vert-bar ()
	 (repeat :for ((ch = (in (:from srm) :char)
			   :then :again)
		       (escaped-char = false :then :again)
		       :collector odd-seg)
;;;;	    (out "(" ch ")")
	    (cond ((char= ch #\\)
		   (!= ch (in (:from srm) :char))
		   (!= escaped-char true)))
;;;;	    (out "^" (:a ch))
	  :until (or (char= ch eof*)
		     (and (not escaped-char) (char= ch #\|)))
	  :result (progn (cond ((char= ch eof*)
				(!= found-end true)))
			 (cons odd-seg (extract-to-colon)))
	  :collect ch))

      ;; Read and discard a char that was already peeked at --
      (:def bite () (in (:from srm) :char))))

;;;;(defvar +import-all+
;;;;      (make-Printable (\\ (srm) (out (:to srm) (:a "+import_all+")))))

;;; Like prefix symbol, but absolutized to a particular Namespace
(def-class Canonized-symbol
   sym
   space
  (:handler
     (print-object (cs srm)
	(cond ((-- +namespace-mode+)
	       (cond (*print-readably*
		      (signal-problem (:class print-not-readable :object cs)))
		     (t
		      (with-slots-from-class (space sym) cs Canonized-symbol
			 (out (:to srm)
			    "#<Canonized symbol " sym " in " space ">")))))
	      (t (out (:to srm) (Canonized-symbol-sym cs)))))))

(defmethod make-load-form ((cs Canonized-symbol) &optional env)
		          (declare (ignore env))
   `(prefixed-symbol-canonicalize ',(Canonized-symbol-sym cs)
				  !()
				  ',(Canonized-symbol-space cs)))

(defun is-CSymbol (x) (or (is-Symbol x) (is-Canonized-symbol x)))

;;; Values 1 and 2 signify greater verbosity
(defvar can-sym-print-verbosity* 0)

(defun can-sym-pprint (srm cs)
   (with-slots-from-class (space sym) cs Canonized-symbol
      (let-fun ()
	 (cond (namespace*
		(multi-let (((path found)
			     (psym-namespace-prefixes !() sym namespace* space)))
		   (out (:to srm)
		     (:q (found
			  (:q ((not (null path)) "@"))
			  (:e (repeat :for ((pre :in path))
				 (:o (:a (string-readable-sym pre))
				     ":")))
			  (:a (string-readable-sym (symbol-name sym))))
			 (t
			  "<no_local_name "
			  (:e (:stream sx)
			    (no-namespace-print sx))
			  " in " (:a (Namespace-name namespace*))
			  ">")))))
	       (t
		(no-namespace-print srm)))
       :where
          (:def no-namespace-print (s)
	     (out (:to s)
		 (:q ((> can-sym-print-verbosity* 1)
		      (:a sym) "<" space)
		     ((> can-sym-print-verbosity* 0)
		      (:a sym) "<" (Namespace-name space))
		     (t
		      "=" (:a sym))))))))

(set-pprint-dispatch 'Canonized-symbol #'can-sym-pprint)
   
;;; Produce a string of characters that would be parsed as a symbol if encountered
;;; by the reader, where the symbol has printname 'str'.
(defun string-readable-sym (str)
   (let ((str-sym (make-symbol str)))
      (subseq (out-to-string str-sym)
	      2)))

(defvar resolve-to-self* !())

;;; Values < symbol, local >.  'local' is true iff 'symbol' comes from 'namespace',
;;; or from a productive ancestor of 'namespace'.
(defun sym-ns-resolve (sym namespace force)
   (cond ((or (not +namespace-mode+)
	      (not sym)
	      (is-Keyword sym)
	      (memq sym resolve-to-self*))
	  (values sym true))
	 (force
	  (prefixed-symbol-canonicalize sym !() namespace))
	 (t
	  (multi-let (((sym _ found-loc)
		       (namespace-lookup namespace sym !() false)))
	     (values sym found-loc)))))

;;; Short-circuit canonicalization by letting second element of each
;;; pair be the canonical version of the first element.  
;;; Second element can be anything, but is probably a Lisp
;;; symbol.
;;; Symbol s by itself is equivalent to (s s).
(defun namespace-associate (ns sym-pairs)
   (cond ((and sym-pairs (atom sym-pairs))
	  (!= sym-pairs (list *-*))))
   (repeat :for ((p :in sym-pairs))
      (cond ((consp p)
	     (let ((name (symbol-as-keyword (car p))))
  		(!= (href (Namespace-table ns) name)
		    (cadr p))))
	    (t
	     (!= (href (Namespace-table ns)
		       (symbol-as-keyword p))
		 p))))
;;;;   (out "sym-pairs = " sym-pairs :%)
;;;;   (dbg-save ns sym-pairs)
;;;;   (breakpoint namespace-associate
;;;;      "Did it work?")
   )

(defun prefixed-symbol-canonicalize (psym prefixes namespace)
   (declare (ignorable namespace))
   (cond ((-- +namespace-mode+)
	  (multi-let (((cs home found-loc)
		       (namespace-lookup namespace psym prefixes true)))
	     (cond (cs
		    (values cs found-loc))
		   (t
		    (!= < psym prefixes > (prefsym-prefixes-separate psym prefixes))
		    (cond ((and home found-loc)
			   (let ((newsym (make-Canonized-symbol psym home)))
			      (!= (href (Namespace-table home) psym)
				  newsym)
			      (values newsym true)))
			  (t
			   (note-defective-exp
			      ((_) "No place to put candidate canonical symbol "
				 (:a (prefixes-symbol->string prefixes psym)))
			      :place prefixed-symbol-canonicalize
			      :fatal)))))))
	 ((null prefixes)
	  (values psym true))
	 (t
	  (note-defective-exp
	     ((_) "Prefixed symbol " (make-Prefixed-symbol prefixes psym)
		  " encountered when not in namespace mode")
	     :place prefixed-symbol-canonicalize
	     :fatal))))

;;; This function implements the rules at the beginning of this file. 
;;; Returns < canonized-symbol, home-namespace, found-as-local >
;;; Third value says whether the canonized-symbol was found in
;;; a productive (wrt 'ns') namespace, 
;;; First value is false if there is no entry for name in the home-namespace.
;;; In this case, second value is the first (and presumably only)
;;; namespace in the ancestor tree that corresponds to the
;;; given prefixes; it's false if there is no productive namespace.
;;; Third value is true if that namespace is productive wrt 'ns'.
;;; If 'local' is false, treat all namespaces but 'ns' itself as unproductive.
(defun namespace-lookup (ns name prefixes local)
   (let-fun ()
      (!= prefixes (remove-empty-strings *-*))
      (cond ((is-Prefixed-symbol name)
	     (!= prefixes (append *-*
				  (remove-empty-strings
				    (Prefixed-symbol-prefix name))))
	     (!= name (Prefixed-symbol-sym name)))
	    (name
	     (!= name (symbol-as-keyword *-*)))
	    (t
	     (signal-problem namespace-lookup
		"Attempt to canonicalize nil")))
      (look-for-it ns prefixes local)

    :where

      (:def look-for-it (ns prefixes local)
;;;;	 (trace-around look-for-it
;;;;	    (:> "(look-for-it: " ns 1 prefixes")")
	 (cond ((null prefixes)
		(let ((can-name
			 (href (Namespace-table ns) name)))
		   (cond (can-name
			  (values can-name ns local))
			 (t
			  (search-parents ns prefixes local)))))
	       (t
		(search-parents ns prefixes local)))
;;;;	    (:< (res-sym nsp fl) "look-for-it: " res-sym 1 nsp 1 fl))
	 )

      (:def search-parents (ns prefixes local)
;;;;	  (trace-around search-parents
;;;;	     (:> "(search-parents: " ns 1 prefixes")")
	  (multi-let (((home found-loc)
		       (cond ((null prefixes)
			      (values ns true))
			     (t
			      (values false false)))))
	     (repeat :for ((par :in (Namespace-parents ns)))
	      :result (values false home found-loc)
	      :within
		 (let ((par-ns (Namespace-import-exporter par))
		       (par-pref (Namespace-import-prefix par))
		       (par-syms (Namespace-import-symbols par)))
;;;;		    (out "Searching " par-ns " pref = " par-pref
;;;;			 " prefixes = " prefixes :%)
		    (let ((exp-syms (Namespace-exports par-ns)))
		       (cond ((or (not par-pref)
				  (and (string= par-pref "")
				       (or (null prefixes)
					   (not (string= (car prefixes) "")))))
			      (cond ((symbol-passes prefixes name exp-syms par-syms)
				     (let ((a (try-parent
						 par-ns prefixes)))
				       (:continue
					:until a
					:result (values a home found-loc))))))
			     ((and (not (null prefixes))
				   (string= par-pref (car prefixes))
				   (symbol-passes (cdr prefixes) name exp-syms par-syms))
			      (let ((a (try-parent
					  par-ns (cdr prefixes))))
				 (:continue
				  :until a
				  :result (values a home found-loc)))))))
	      :where

		(:def try-parent (par-ns prefixes)
;;;;		   (trace-around try-parent
;;;;		      (:> "(try-parent: " par-ns 1 prefixes")")
		   (let ((par-creatable (and local
					     (Namespace-import-productive par))))
		      (multi-let (((a h fl)
				   (look-for-it
				      par-ns prefixes par-creatable)))
			 (cond ((or a
				    (and h
					 (or (not home)
					     (and fl (not found-loc)))))
				(!= home h)
				(!= found-loc fl)))
			 a))
;;;;		      (:< (val &rest _) "try-parent: " val))
		   )))
;;;;	     (:< (res-sym nsp fl) "search-parents: " res-sym 1 nsp 1 fl))
	  )))

;;; If symbol is located in parent with the given exports, will it pass
;;; to child through the given imports?  
;;; We've already checked that the prefix that goes with these imports
;;; is appropriate; the 'prefixes' variable is the remaining prefixes.       
;;; We ignore the possibility that the same Canonized-symbol could be known
;;; by multiple paths from the exporter.       
(defun symbol-passes (prefixes sym exports imports)
   (let-fun ()
      (and (passes-through exports)
	   (passes-through imports))
    :where
       (:def passes-through (spec)
	   (or (eq spec +all-symbols+)
	       (cond ((car-eq spec +all-symbols-but+)
		      (not (in-list (cdr spec))))
		     (t (in-list spec)))))

       (:def in-list (psyms)
	   (exists (ps :in psyms)
	      (and (prefix-lists-equal (Prefixed-symbol-prefix ps)
				       prefixes)
		   (eq (Prefixed-symbol-sym ps)
		       sym))))))

(defun namespace-add-import (ns import-spec)
   (let ((prev-spec (existing-import-spec ns import-spec))
	 (new-super (Namespace-import-exporter import-spec)))
      (dbg-save ns import-spec prev-spec new-super)
;;;;      (breakpoint namespace-import-exporter
;;;;         "here")
      (cond ((not new-super)
	     (dbg-save ns import-spec)
	     (signal-problem namespace-add-import
		"Null exporter in import-spec " import-spec))
	    ((eq new-super ns)
	     (dbg-save ns)
	     (signal-problem namespace-add-import
		"About to link " ns " to itself"))
	    ((exists (p :in (Namespace-parents new-super))
		(eq (Namespace-import-exporter p) ns))
	     (signal-problem namespace-add-import
		"About to set up namespace-import cycle between"
		:% 2 ns " and " new-super))
	    ((and prev-spec
                  (not (eq (Namespace-import-productive prev-spec)
                           (Namespace-import-productive import-spec))))
	     (err-out "Warning -- adding import spec with productivity change"
		      1 (Namespace-import-productive prev-spec)
		        " --> " (Namespace-import-productive import-spec)
		      :% " Importer: " ns
		      :% " New import spec: " import-spec)))
      (let ((bdgs
	       (<# (\\ (psym)
		      (tuple psym (namespace-lookup new-super psym !() false)))
		   (cond (prev-spec (new-imports import-spec prev-spec))
			 (t (import-exported-syms import-spec))))))
         (dbg-save bdgs)
	 (let ((bad-ones
		  (bdgs-check-for-conflicts
		      bdgs ns import-spec)))
	    (cond ((null bad-ones)
		   (let ((new-productive-ancs
			    (all-productive-ancestors import-spec))
			 (old-productive-ancs
			    (<! all-productive-ancestors
				(Namespace-parents ns))))
		      (let ((ambig-cases
			       (repeat :for ((n :in new-productive-ancs))
				:nconc
				  (repeat :for ((o :in old-productive-ancs))
				   :when (and (equal (first n) (first o))
					      (not (eq (second n) (second o))))
				   :collect (tuple n o)))))
			 (cond ((not (null ambig-cases))
				(note-defective-exp
				   ((ns)
				    "Adding import-spec"
				    " to " ns
				    :% " will make symbol prefix(es) ambiguous:"
				    (:e (repeat :for ((amb :in ambig-cases))
					   (:o "Prefix "
					       (:e (repeat
						       :for ((pref
							      :in (first (first amb))))
						     (:o (:a pref) ":")))
					       :% " New designee: " (second (first amb))
					       :% " Old desginee: " (second (second amb))
					       :%))))
				   :target ns
				   :place namespace-add-import
				   (:proceed
				    "I'll assume the ambiguity is unimportant"))))))
		   (!= (Namespace-children new-super)
		       (adjoinq ns *-*))
		   (!= (Namespace-parents ns)
		       (cons import-spec
			     (cond (prev-spec (delete prev-spec *-* :test #'eq :count 1))
				   (t *-*)))))
		  (t
		   (note-defective-exp
		      ((ns)
		       "Adding import " import-spec
		       " to " ns
		       :% " will cause conflicts at " :%
		       (:e (:stream e) (conflicts-out bad-ones e)))
		      :target ns
		      :place namespace-add-import
		      :fatal)))))))

(defun existing-import-spec (ns new-spec)
   (let ((str (Namespace-import-prefix new-spec))
	 (super-ns (Namespace-import-exporter new-spec)))
      (namespace-has-import ns str super-ns)))

(defun namespace-has-import (ns pref exporter)
      (repeat :for ((par :in (Namespace-parents ns)))
       :result false
       :until (and (cond (pref
			  (and par
			       (string= (Namespace-import-prefix par)
					pref)))
			 (t (not (Namespace-import-prefix par))))
		   (eq (Namespace-import-exporter par)
		       exporter))
       :result par))

;;; Find all the ancestor namespaces of Namespace-import 'ns-imp' into
;;; which new symbols can be generated.  
;;; Return as list of pairs (prefixes-list namespace).
;;; 'prefixes-list' is sub-to-super order.
(defun all-productive-ancestors (ns-imp)
   (let ((ns (Namespace-import-exporter ns-imp))
	 (prefix (Namespace-import-prefix ns-imp)))
      (let ((ns-prod-ancestors
	       (repeat :for ((par :in (Namespace-parents ns)))
		:nconc
		   (let ((ancs-short-prefix (all-productive-ancestors par)))
		      (cond ((and prefix (not (string= prefix "")))
			     (<# (\\ (p)
				    (tuple (cons prefix (first p))
					   (second p)))
				 ancs-short-prefix))
			    (t
			     ancs-short-prefix))))))
	 (cond ((Namespace-import-productive ns-imp)
		(cons (tuple (cond (prefix (list prefix))
				   (t !()))
			     ns)
		      ns-prod-ancestors))
	       (t
		ns-prod-ancestors)))))

;;; Obviously we can't check all the children, because domains anywhere
;;; on the web could have used this namespace.
(defun namespace-change-exports (namespace old-exports new-exports)
   (repeat :for ((child :in (Namespace-children namespace)))
      (repeat :for ((imp :in (Namespace-parents child)))
       :result (signal-problem namespace-change-exports
		  "Fumbled Namespace-import for parent " namespace
		  " of " child
		  :proceed)
       :until (eq (Namespace-import-exporter imp)
		  namespace)
       :result (let ((new-syms (new-export-new-syms
				   new-exports old-exports imp)))
		  (let ((bdgs
			   (<# (\\ (psym)
				  (tuple psym (namespace-lookup namespace psym
								!() false)))
			       new-syms)))
		     (let ((bad-ones (bdgs-check-for-conflicts
					bdgs namespace imp)))
			(cond ((null bad-ones)
			       (!= (Namespace-exports namespace)
				   new-exports))
			      (t
			       (note-defective-exp
				   ((ns)
				    "Changing exports of " ns
				    " to " new-exports
				    :% " will cause conflicts at " :%
				    (:e (:stream e) (conflicts-out bad-ones e)))
				   :target namespace
				   :place namespace-change-exports
				   :fatal)))))))))

(defun bdgs-check-for-conflicts (bdgs ns import-spec)
   ;;; --bdgs should be a list of tuples < prefixes, canonized-symb >	 
   ;;;;	 (out "bdgs = " bdgs :%)
	 ;; bdgs is list of (Prefixed-symbol val) pairs.
	 (let-fun ((check-descendents (descendent bdgs path)
				       ;; -- descendent is a Namespace
;;;;		      (trace-around check-descendents
;;;;			 (:> "(check-descendents: " descendent 1 bdgs 1 path ")")
		      (cond ((null bdgs) !())
			    (t
			     (repeat :for ((bdg :in bdgs)
					   :collector bad-ones)
			      :within
			         (multi-let (((a _ _)
					      (namespace-lookup descendent
								(first bdg)
								path false)))
				    (cond ((and a
						(not (eql a (second bdg))))
;;;;					   (dbg-save a bdg descendent bdgs
;;;;						     ns import-spec prev-spec
;;;;						     new-super path)
					   (:continue
					    :collect (tuple descendent bdg a)))))
			      :result
				 (nconc bad-ones
					(repeat :for ((ch :in (Namespace-children
								 descendent)))
					   (repeat :for ((chi :in (Namespace-parents
								     ch)))
					    :when (eq (Namespace-import-exporter
							 chi)
						      descendent)
					    :nconc
					       (check-descendents
						  ch
						  (bdgs-visible bdgs path chi descendent)
						  (namespace-import-prefix-cons
						     chi path))))))))
;;;;			 (:< (&rest _) "check-descendents: exit" ))
		      ))
	    (check-descendents
	       ns bdgs
	       (namespace-import-prefix-cons import-spec !()))

	  :where

	    (:def bdgs-visible (bdgs path imspec parent)
	       (repeat :for ((bdg :in bdgs))
		:when (let ((exportsyms (Namespace-exports parent))
			    (importsyms (Namespace-import-symbols imspec)))
			 (and (or (eq exportsyms +all-symbols+)
				  (cond ((car-eq exportsyms +all-symbols-but+)
					 (not (bdg-in-psym-list
						 bdg path (cdr exportsyms))))
					(t
					 (bdg-in-psym-list bdg path exportsyms))))
			      (or (eq importsyms +all-symbols+)
				  (cond ((car-eq importsyms +all-symbols-but+)
					 (not (bdg-in-psym-list
					         bdg path (cdr importsyms))))
					(t
					 (bdg-in-psym-list
					    bdg path importsyms))))))
		:collect bdg))

	    (:def bdg-in-psym-list (bdg psym-list path)
	       (let ((sym (Prefixed-symbol-sym (car bdg)))
		     (pref (Prefixed-symbol-prefix (car bdg))))
		  (exists (psym :in psym-list)
		     (and (eq (Prefixed-symbol-sym psym)
			      sym)
			  (equal-app (Prefixed-symbol-prefix psym)
				     path pref)))))

	    ;; List of strings l is equal to (append l1 l2)
	    (:def equal-app (l l1 l2)
	        (cond ((null l1) (equal l l2))
		      ((null l) false)
		      (t
		       (and (string= (car l) (car l1))
			    (equal-app (cdr l) (cdr l1) l2)))))))

(defun conflicts-out (bad-ones out-stream)
   (out (:to out-stream)
      (:e (repeat :for ((bad :in bad-ones))
	     (:o "Namespace: " (first bad)
		 " Symbol: " (first (second bad))
		 " Old resolution: " (second (second bad))
		 " New: " (third bad))))))
					      
;;; Returns list of Prefixed-symbols that will be newly inherited
;;; when new-spec replaces old version of link between the same two
;;; namespaces.
(defun new-imports (new-spec old-version)
   (let ((old-syms (Namespace-import-symbols old-version))
	 (new-syms (Namespace-import-symbols new-spec)))
      (cond ((eq old-syms +all-symbols+)
	     !())
	    (t
	     (let ((imp-syms (import-exported-syms new-spec)))
		(pref-sym-list-intsect
	           imp-syms
		   (cond ((car-eq old-syms +all-symbols-but+)
			  (cond ((eq new-syms +all-symbols+)
				 (cdr old-syms))
				((car-eq new-syms +all-symbols-but+)
				 ;; old: (- a b) + new: (- b c) -> (a)
				 (pref-sym-list-diff (cdr old-syms) (cdr new-syms)))
				(t
				 ;; old: (- a b c) + new: (c d) -> (c d a b)
				 (append new-syms
					 (pref-sym-list-diff
					    (cdr old-syms)
					    new-syms)))))
			 (t
			  (pref-sym-list-diff new-syms old-syms)))))))))

(defun new-export-new-syms (new-exports old-exports import-spec)
   (cond ((eq old-exports +all-symbols+)
	  !())
	 (t
	  (let ((exporting-namespace
		    (Namespace-import-exporter import-spec)))
	     (set-difference
		(namespace-exports-symbols exporting-namespace
					   new-exports)
		(namespace-exports-symbols exporting-namespace
					   old-exports))))))

;;; Return list of prefixed syms that make it through the filters from
;;; the export side to the import side of 'import-spec'.
(defun import-exported-syms (import-spec)
   (let ((imp-syms (Namespace-import-symbols import-spec))
	 (exported-syms (namespace-import-all-exports import-spec)))
      (cond ((eq imp-syms +all-symbols+)
	     exported-syms)
	    ((car-eq imp-syms +all-symbols-but+)
	     (pref-sym-list-diff exported-syms (cdr imp-syms)))
	    (t
	     (pref-sym-list-intsect imp-syms exported-syms)))))

;;; Return list of all symbols that make it out the exporter of 'import-spec',
;;; usually acquiring a new prefix.
(defun namespace-import-all-exports (import-spec)
   (let ((import-spec-exports (namespace-all-exports
				 (Namespace-import-exporter import-spec)))
	 (import-spec-pref (Namespace-import-prefix import-spec)))
      (cond ((or (not import-spec-pref)
		 (string= import-spec-pref ""))
	     import-spec-exports)
	    (t
	     (<# (\\ (psym)
		    (make-Prefixed-symbol
		       (cons import-spec-pref (Prefixed-symbol-prefix psym))
		       (Prefixed-symbol-sym psym)))
		 import-spec-exports)))))

;;; Return list of all prefixed symbols that are exported from 'ns'.
(defun namespace-all-exports (ns)
   (namespace-exports-symbols ns (Namespace-exports ns)))

(defun namespace-exports-symbols (ns exports)
   (with-slots-from-class (parents table) ns Namespace
      (cond ((or (eq exports +all-symbols+)
		 (car-eq exports +all-symbols-but+))
	     (let ((all-cands
		      (nconc (<# (\\ (ksym)
				    (make-Prefixed-symbol (list "") ksym))
				 (hash-table-all-keys table))
			     (<! namespace-import-all-exports
				 parents))))
		(cond ((eq exports +all-symbols+)
		       all-cands)
		      (t
		       (pref-sym-list-diff all-cands (cdr exports))))))
	    (t
	     (list-copy exports)))))

(defun pref-sym-list-diff (psl1 psl2)
   (repeat :for ((ps :in psl1))
    :when (not (psym-in-psym-list ps psl2))
    :collect ps))

(defun pref-sym-list-intsect (psl1 psl2)
   (repeat :for ((ps :in psl1))
    :when (psym-in-psym-list ps psl2)
    :collect ps))

(defun psym-in-psym-list (ps psl)
   (exists (ps1 :in psl)
      (and (prefix-lists-equal (Prefixed-symbol-prefix ps1)
			       (Prefixed-symbol-prefix ps))
	   (eq (Prefixed-symbol-sym ps1)
	       (Prefixed-symbol-sym ps)))))

(defun prefix-lists-equal (sl1 sl2)
   (!= sl1 (remove "" *-* :test #'string=))
   (!= sl2 (remove "" *-* :test #'string=))
   (and (= (length sl1) (length sl2))
	(<& string= sl1 sl2)))

(defun local-sym (symish local-namespace &optional (pkg *package*))
   (multi-let (((ln found)
		(local-name symish local-namespace true)))
      (cond (found
	     (cond ((is-Symbol ln) ln)
		    (t
		     (intern ln pkg))))
	    (t
	     (signal-problem local-sym
		"No local name found for " symish " in " local-namespace)))))

;;; Returns < local-name, true> or <nil, false> if there is no local name.
;;; 'local-name' is a string
;;; If 'must-find' is true, the second case signals an error.
(defun local-name (symish local-namespace must-find)
   (cond ((is-Symbol symish) (values symish true))
	 ((is-Canonized-symbol symish)
	  (let ((home-namespace (Canonized-symbol-space symish))
		(sym (Canonized-symbol-sym symish)))
	     (multi-let (((path found)
			  (psym-namespace-prefixes
			     !() (Canonized-symbol-sym symish)
			     local-namespace home-namespace)))
	        (cond (found
		       (cond ((null path)
			      (values (symbol-name sym)
				      true))
			     (t
			      (values (prefixes-symbol->string path sym)))))
		      (t
		       (cond (must-find
			      (bind ((namespace* false)
				     (can-sym-print-verbosity* 2))
				 (signal-problem local-name
				    "No local name for canonized symbol "
				    symish " in "  local-namespace
				    (:proceed
				     "I'll report to caller that there is no local name")))))
		      (values nil false))))))
	 (t
	 (cond (must-find
		(signal-problem local-name
		   "Can't compute local name for " symish
		   (:proceed "I'll report to caller that there is no local name"))))
	 (values nil false))))

;;; 'sym' is a keyword
(defun psym-namespace-prefixes (prefixes sym low-ns super-ns)
   (cond ((eq low-ns super-ns)
	  (values prefixes true))
	 (t
	  (repeat :for ((ch :in (Namespace-children super-ns)))
	   :result (values nil false)
	   :within
	     (let ((imp (namespace-psym-link ch prefixes sym super-ns)))
	        (multi-let (((end-prefs ok)
			     (psym-namespace-prefixes
				(namespace-import-prefix-cons imp prefixes)
				sym low-ns ch)))
		   (:continue
		    :until ok
		    :result (values end-prefs true))))))))

(defun namespace-psym-link (sub-ns prefixes sym super-ns)
   (repeat :for ((imp :in (Namespace-parents sub-ns)))
    :result (signal-problem namespace-psym-link
	       "Fumbled namespace link for "
	       (:a (prefixes-symbol->string prefixes sym))
	       " from " sub-ns " to " super-ns)
    :until (and (eq (Namespace-import-exporter imp)
		    super-ns)
		(symbol-passes prefixes sym
			       (Namespace-exports super-ns)
			       (Namespace-import-symbols imp)))
    :result imp))

(defun prefixes-symbol->string (path sym)
   (out-to-string
      (:q ((forall (p :in path) (string= p ""))
	   (:a sym))
	  (t
	   "@"
	   (:e (repeat :for ((pre :in path))
		  (:o (:q ((not (string= pre ""))
			   (:a pre) ":")))))
	   (:a sym)))))

;;; "Normalize" by shifting all of the prefixes off 'psym' and on to 
;;; 'prefixes'.  Return resulting <psym, prefixes>.
(defun prefsym-prefixes-separate (name prefixes)
   (let-fun ()
      (!= prefixes (remove-empty-strings *-*))
      (cond ((is-Prefixed-symbol name)
	     (!= prefixes (append *-*
				  (remove-empty-strings
				    (Prefixed-symbol-prefix name))))
	     (!= name (Prefixed-symbol-sym name)))
	    (name
	     (!= name (symbol-as-keyword *-*)))
	    (t
	     (signal-problem prefsym-prefixes-separate
		"Attempt to canonicalize nil")))
      (values name prefixes)))

(defun remove-empty-strings (l) (remove "" l :test #'string=))

;;;; Not used because the validity of a prefix chain depends on the symbol 
;;;; it will be the prefix of.
#|
;;; Find list of prefix strings linking super-ns to low-ns
;;; Return < list, found-chain >,
;;; where 'found-chain' is false if no such list can be found, else true.
(defun namespace-prefixes (low-ns super-ns)
   (let ((checked !())) ; List of Namespaces checked so far
      (let-fun ((anc-search (ns)
;;;;		   (trace-around anc-search
;;;;		      (:> "(anc-search: " ns ")")
		   (cond ((eq ns super-ns)
			  (values !() true))
			 ((memq ns checked)
			  (values nil false))
			 (t
			  (!= checked (cons ns *-*))
;;;;			  (cond ((and (null (Namespace-parents ns))
;;;;				      (not (eq ns opt::requirement-namespace*))
;;;;				      (not (eq (Namespace-name ns) 'opt::opt)))
;;;;				 (dbg-save :run-loud ns low-ns super-ns)
;;;;				 (breakpoint namespace-prefixes
;;;;				    "Namespace with no parents: " ns)))
			  (repeat :for ((p :in (Namespace-parents ns))
					(found false)
					shortest (shortest-length 10000))
			   :within
			     (multi-let (((chn found-now)
					  (anc-search (Namespace-import-exporter p))))
				(:continue
				   (cond (found-now
					  (!= found true)
;;;;					  (out "found " chn :%)
					  (let ((pref (Namespace-import-prefix p)))
					     (let ((preflen (length pref)))
					        (let ((chnlen
						         (cond ((= preflen 0)
								(length chn))
							       (t
								(+ (length chn) 1)))))
					           (cond ((< chnlen shortest-length)
							  (!= shortest-length chnlen)
							  (!= shortest
							      (cond ((= preflen 0)
								     chn)
								    (t
								     (cons pref
									   chn)))))))))))
				 :until (= shortest-length 0)))
			   :result (values shortest found))))
;;;;		      (:< (pl found) "anc-search: " pl 1 found))
		   ))
	 (multi-let (((chn found)
		      (anc-search low-ns)))
;;;;	    (cond (found
;;;;		   (out "returning " chn :%)))
	    (values chn found)))))
|#

(defun symbol-as-keyword (s)
   (cond ((is-Keyword s) s)
	 ((is-Symbol s)
	  (intern (Symbol-name s) keyword-package*))
	 (t
	  (signal-problem symbol-as-keyword
	     "Can't move non-symbol to keyword package: " s))))

(defun hash-table-all-keys (ht)
   (let ((keys !()))
      (maphash (\\ (k _) (!= keys (cons k *-*)))
	       ht)
      keys))

(defun string-split (s char)
   (declare (string s) (character char))
   (let ((l (length s)))
      (repeat :for ((nextpos = (or (position char s) l)
			     :then (cond ((< nextpos l)
					  (or (position char s
							:start (+ nextpos 1))
					      l))
					 (t l)))
		    (prevpos 0 :then (+ nextpos 1))
		    :collector substrings)
       :until (>= prevpos l)
       :when (not (= nextpos prevpos))
       :collect (subseq s prevpos nextpos)
       :result substrings)))

(defun is-symbolish (x)
   (or (is-Symbol x)
       (cond ((or (is-Prefixed-symbol x)
		  (is-Canonized-symbol x))
	      (or (-- +namespace-mode+)
		  (bind ((namespace* false))
		     (signal-problem is-symbolish
			"Encountered " x " in non-namespace mode"))))
	     (t false))))


(defun no-namespace-add (l1 l2)
   (declare (ignorable l1))
   (cond ((-- +namespace-mode+) l2)
	 (t (append l1 l2))))

;;; For debugging
(defun namespace-show (ns)
   (with-hash-table-iterator (ns-iter (Namespace-table ns))
      (bind ((can-sym-print-verbosity* 1))
	 (repeat 
	  :within
	    (multiple-value-let (found key value)
				(ns-iter)
	       (:continue
		:while found
		   (out key " -- " value :%)))))))
	    