;-*- Mode: Common-lisp; Package: ytools; Readtable: ytools; -*-
(in-package :ytools)
;;; $Id: symboid.lisp,v 1.5 2006/11/08 01:17:03 dvm Exp $

(depends-on %module/ ytools)

(needed-by-macros
   (export '(make-Symboid is-Symboid Symboid-name symboid-get
	     build-symboid Symboid intern-printed-wheresyms*)))

(defpackage wheresym-read)

(defvar wheresym-package* (find-package ':wheresym-read))

(defvar intern-printed-wheresyms* true)

;; Wheresyms behave something like gensyms that are interned when printed.
(defstruct (Wheresym
               (:print-function
		   (lambda (y srm k)
		       (declare (ignore k))
                       (let ((sym (where->here y)))
                         (cond (intern-printed-wheresyms*
                                (bind ((*package* wheresym-package*))
                                   (out (:to srm) "!:" sym)))
                               (t
                                (let ((str (out-to-string sym)))
                                   (cond ((eql (mismatch "#:" str)
                                               2)
                                          (out (:to srm)
                                               "!"
                                               (:a (subseq str 1))))
                                         (t
                                          (out (:to srm)
                                               "!:"
                                               (:a str))))))))))
	       (:constructor
		   make-Wheresym (name))
	       (:predicate is-Wheresym))
   (name "" :type string)
   (sym nil)
   (plist '()))

(defun wheresym-get (s ind)
  (let ((p (assoc ind (Wheresym-plist s) :test #'eq)))
    (cond (p (cdr p))
	  (t nil))))

(defun wheresym-put (s ind val)
  (let ((p (assoc ind (Wheresym-plist s) :test #'eq)))
    (cond ((null p)
	   (setf p (cons ind val))
	   (on-list p (Wheresym-plist s)))
	  (t
	   (setf (cdr p) val)))
    val))

(defsetf wheresym-get wheresym-put)

;; If a Wheresym is saved to a file, then it's crucial when restoring
;; that it be 'eq' to the other items restored from the same Wheresym.
;; The problem is that the obvious way of ensuring that is not guaranteed
;; to work.  The obvious method is to intern the Wheresym, then use
;; the 'eq'uality of the attached symbols to establish the guarantee.
;; The bug is that the environment when we restore might already have
;; a Wheresym with that symbol.  (Indeed, it's very likely to if
;; symbols are generated using counters from 0 in the usual way.)
;; So we have to grab the symbol away from any competing Wheresym that
;; has it at restore time.  How do we know it's a competitor and not
;; a previously restored copy of the Wheresym?
;; We use a bizarre property with a random value to discriminate.

(defparameter distinguisher-prop*
    (build-symbol (:package :wheresym-read) distinguisher))

(defmethod make-load-form ((wh Wheresym) &optional _)
   (cond ((not intern-printed-wheresyms*)
          (signal-problem Wheresym/make-load-form
             "Can't save uninterned Wheresyms in compiled files")))
   (let ((distinguisher
             (wheresym-get wh distinguisher-prop*)))
      (cond ((not distinguisher)
             (!= distinguisher (random 100000))
             (!= (wheresym-get wh distinguisher-prop*)
                 distinguisher)))
      (values
         `(restore-wheresym ',(Wheresym-name wh) ',distinguisher)
         `(progn
             ,@(repeat :for ((e :in (Wheresym-plist wh)))
                :collect
                  `(!= (wheresym-get ',wh ',(first e))
                       ',(rest e)))))))

(defun restore-wheresym (name distinguisher)
   (cond ((not intern-printed-wheresyms*)
          (signal-problem restore-wheresym
             "In order to restore Wheresyms from compiled files,"
             :% " the variable 'intern-printed-wheresyms*' must be set to true."
             (:proceed "I will set 'intern-printed-wheresyms*' to true"))
          (!= intern-printed-wheresyms* true)))
   (let* ((wh-sym (intern name wheresym-package*))
          (existing (get wh-sym 'wheresym)))
      (let-fun ()
         (cond (existing
                (cond ((eql (wheresym-get existing distinguisher-prop*)
                            distinguisher)
                       existing)
                      (t
                       ;; Change the existing one's name and symbol
                       (let ((ex-new (wheresym-find-another-name
                                         existing wh-sym)))
                          (!= (get ex-new 'wheresym) existing)
                          (!= (Wheresym-sym existing) ex-new))
                       (restore-to-wh-sym))))
               (t
                (restore-to-wh-sym)))

       :where

         (:def restore-to-wh-sym ()
            ;; I think it's important to get this property restored
            ;; now, even though all the properties will be restored
            ;; by the "initialization form" returned by 'make-load-form'.
            (let ((new-wh (create-wheresym-and-sym-link name wh-sym)))
               (!= (wheresym-get new-wh distinguisher-prop*)
                   distinguisher)
               new-wh)))))


(defun where->here (wh)
   (let-fun ((:def sym-name ()
                (Wheresym-name wh))
             )

      (let ((s (Wheresym-sym wh)))
         (cond ((not s)
                (setf s (>< (cond (intern-printed-wheresyms*
                                   (\\ (s) (intern s wheresym-package*)))
                                  (t
                                   #'make-symbol))
                            (sym-name)))
                (cond ((get s 'wheresym)
                       (!= s (wheresym-find-another-name wh s))
                       (err-out "Changing symbol associated with wheresym"
                                " to " s)))
                (!= (get s 'wheresym) wh)
                (!= (Wheresym-sym wh) s))
               ((and intern-printed-wheresyms*
                     (not (symbol-package s)))
                ;; Policy changed
                (let ((s1 (intern (sym-name) wheresym-package*)))
                   (cond ((get s1 'wheresym)
                          (!= s1 (wheresym-find-another-name wh s1))
                          (err-out "Changing symbol associated with wheresym"
                                   " to " s1)))
                   (!= (get s1 'wheresym) wh)
                   (!= (Wheresym-sym wh) s1)))
               ((and (not intern-printed-wheresyms*)
                     (symbol-package s))
                (err-out "Uninterning stray wheresym with sym " s)
                (unintern s wheresym-package*)
                s)
               (t s)))))

(defun wheresym-find-another-name (wh sym)
                (repeat :for ((i = 1 :to 1000)
                              (str (Symbol-name sym)))
                 :result (signal-problem wheresym-find-another-name
                            "Boy are there a lot of bogus wheresyms out there")
                 :within
                    (let* ((str1 (out-to-string (:a str) "-" i))
                           (s (intern str1 wheresym-package*)))
                       (:continue
                        :until (not (get s 'wheresym))
                        :result (progn
                                   (!= (Wheresym-name wh) str1)
                                   s)))))

(defun set-wheresym-interning (val)
   (!= intern-printed-wheresyms*
       (cond ((or (not val) (eq val ':off))
              (do-symbols (sym wheresym-package*)
                 (unintern sym wheresym-package*))
              false)
             (t
              true))))

;;;;(defun string->wheresym (s)
;;;;   (let ((sym (intern (concatenate 'string "!:" s))))
;;;;      (or (get sym 'wheresym)
;;;;	  (error-break string->wheresym
;;;;	     "Unknown symboid " sym
;;;;	     (:prompt-for "Intended symboid")))))

;;;; (defvar wheresym-read-package* (find-package ':wheresym-read))

(defvar create-wheresyms-when-read* true)

(def-excl-dispatch #\: (srm _)
   (let ((x (bind ((*package* wheresym-package*))
	       (read srm))))
      (cond ((and (is-Symbol x)
                  (eq (symbol-package x) wheresym-package*))
;;;;	     (unintern x wheresym-read-package*)
             (let ((wh (get x 'wheresym)))
                (cond (wh wh)
                      ((or create-wheresyms-when-read*
                           (progn
                              (signal-problem excl-colon-reader
                                 "Unknown wheresym !:" x
                                 (:proceed "I'll create it"))
                              true))
                       (create-wheresym-and-sym-link
                           (Symbol-name x)
                           x)))))
	    (t
	     (signal-problem colon-exmac
		    "Ill-formed wheresym !:" x
		    (:prompt-for "Intended symboid"))))))

;;;; (set-dispatch-macro-character #\! #\: #'colon-exmac ytools-readtable*)

(defun create-wheresym-and-sym-link (name sym)
   (let ((wh (make-Wheresym name)))
      (!= (Wheresym-sym wh) sym)
      (!= (get sym 'wheresym) wh)
      wh))

(cl:deftype Symboid ()
   '(or symbol Wheresym))

(defun make-Symboid (str) (make-Wheresym str))

(defun is-Symboid (s) (or (symbolp s) (is-Wheresym s)))

(defun Symboid-name (s)
      (cond ((is-Wheresym s) (Wheresym-name s))
	    (t (symbol-name s))))

(defun symboid-get (s ind)
      (cond ((is-Wheresym s)
	     (wheresym-get s ind))
	    (t
	     (get s ind))))

(defun symboid-put (s ind val)
     (cond ((is-Wheresym s)
	    (setf (wheresym-get s ind) val))
	   (t
	    (setf (get s ind) val))))

(defsetf symboid-get symboid-put)

(defmacro build-symboid (&rest l)
   `(make-Symboid ,(symstuff l)))

;;; This is to make sure that the symbol gets interned
'Symboid
